//	$Id: devif.cpp,v 1.1 1999/08/26 08:04:38 cisc Exp $

#include "headers.h"
#include "cdif.h"
#include "systemif.h"
#include "pc88/pcinfo.h"

#define EXTDEVAPI	__declspec(dllexport)

using namespace PC8801;

struct Holder
{
	SystemIF* sys;
	CDIF* cdif;
};

enum SpecialPort
{
	pint0 = 0x100, 
	pint1, pint2, pint3, pint4, pint5, pint6, pint7,
	pres,			// reset
	pirq,			// IRQ
	piack,			// interrupt acknowledgement
	vrtc,			// vertical retrace
	popnio,			// OPN ̓o̓|[g 1
	popnio2,		// OPN ̓o̓|[g 2 (A)
	portend
};

// ---------------------------------------------------------------------------

const int OutTable[] =
{
	pres, 0x90, 0x91, 0x94, 0x97, 0x98, 0x99, 0x9f, -1,
};
const int InTable[] =
{
	0x90, 0x91, 0x92, 0x93, 0x96, 0x98, 0x99, 0x9b, 0x9d, -1,
};

// ---------------------------------------------------------------------------
//	outporttable Ɏw肵|[gɑ΂ Out 삪sꂽɌĂ΂
//
void OutPort(void* h, uint port, uint data)
{
	Holder* hld = reinterpret_cast<Holder*> (h);
	switch (port)
	{
	case pres:
		hld->cdif->Enable((data & 0x40) != 0);
		break;

	case 0x90:	hld->cdif->Out90(port, data);	break;
	case 0x91:	hld->cdif->Out91(port, data);	break;
	case 0x94:	hld->cdif->Out94(port, data);	break;
	case 0x97:	hld->cdif->Out97(port, data);	break;
	case 0x98:	hld->cdif->Out98(port, data);	break;
	case 0x99:	hld->cdif->Out99(port, data);	break;
	case 0x9f:	hld->cdif->Out9f(port, data);	break;
	}
}

// ---------------------------------------------------------------------------
//	inporttable Ɏw肵|[gɑ΂ In 삪sꂽɌĂ΂
//
uint InPort(void* h, uint port)
{
	Holder* hld = reinterpret_cast<Holder*> (h);
	switch (port)
	{
	case 0x90:	return hld->cdif->In90(port); break;
	case 0x91:	return hld->cdif->In91(port); break;
	case 0x92:	return hld->cdif->In92(port); break;
	case 0x93:	return hld->cdif->In93(port); break;
	case 0x96:	return hld->cdif->In96(port); break;
	case 0x98:	return hld->cdif->In98(port); break;
	case 0x99:	return hld->cdif->In99(port); break;
	case 0x9b:	return hld->cdif->In9b(port); break;
	case 0x9d:	return hld->cdif->In9d(port); break;
	}
	return 0xff;
}

// ---------------------------------------------------------------------------
//	ԕۑEAs
//	data == 0	: ԕۑf[^̃oCgԂ
//	data != 0	:
//		save == 0	: ԕA
//		save != 0	: ԕۑ
//
uint SnapShot(void* h, uint8* data, bool save)
{
	CDIF* cdif = reinterpret_cast<Holder*> (h)->cdif;
	if (!data)
	{
		return cdif->GetStatusSize();
	}
	else
	{
		if (save)
			return cdif->SaveStatus(data);
		else
			return cdif->LoadStatus(data);
	}
}

extern "C"
{

//	SystemIF  CDIF 쐬
EXTDEVAPI void* __cdecl Connect
(void* h, const PCInfo* p, DeviceInfo* di)
{
	if (di->rev != DeviceInfo::direv || di->size != sizeof(DeviceInfo))
		return 0;

	Holder* holder = new Holder;
	if (!holder)
		return 0;

	holder->sys = new SystemIF(h, p);
	holder->cdif = new CDIF(DEV_ID('c', 'd', 'i', 'f'));
	if (!holder->sys || !holder->cdif || !holder->cdif->Init(holder->sys))
	{
		delete holder->sys;
		delete holder->cdif;
		delete holder;
		return 0;
	}

	di->flags = 0;
	di->id = holder->cdif->GetID();
	di->inporttable = InTable;
	di->outporttable = OutTable;
	di->inport = InPort;
	di->outport = OutPort;
	di->snapshot = SnapShot;

	holder->cdif->Enable(true);

	return reinterpret_cast<void*> (holder);
}

//	j
EXTDEVAPI bool __cdecl Disconnect(void* h)
{
	Holder* holder = reinterpret_cast<Holder*> (h);
	if (holder)
	{
		delete holder->sys;
		delete holder->cdif;
		delete holder;
		return true;
	}
	return false;
}

}

BOOL APIENTRY DllMain(HANDLE hmod, DWORD rfc, LPVOID)
{
	switch (rfc)
	{
	case DLL_PROCESS_ATTACH:
		break;
	
	case DLL_THREAD_ATTACH:
		break;
	
	case DLL_THREAD_DETACH:
		break;
	
	case DLL_PROCESS_DETACH:
		break;
	}
	return true;
}

