// ---------------------------------------------------------------------------
//	M88 - PC-8801 Emulator
//	Copyright (C) cisc 1997, 1999.
// ---------------------------------------------------------------------------
//	$Id: opnif.cpp,v 1.14 1999/08/23 11:58:53 cisc Exp $

#include "headers.h"
#include "types.h"
#include "misc.h"
#include "schedule.h"
#include "pc88/sound.h"
#include "pc88/opnif.h"
#include "pc88/config.h"
#include "status.h"

#define LOGNAME "opnif"
#include "diag.h"

using namespace PC8801;

// ---------------------------------------------------------------------------
//	vXP[̐ݒl
//	static ɂ̂́CFMGen ̐ɂC OPN قȂNbN
//	邱ƂoȂ߁D
//
int OPNIF::prescaler = 0x2d;

// ---------------------------------------------------------------------------
//	Ej
//
OPNIF::OPNIF(const ID& id, int i)
: Device(id)
{
	scheduler = 0;
	sound = 0;
	srate = 0;
	imaskbit = i;
	opnamode = false;
	nextcount = 0;
}

OPNIF::~OPNIF()
{
}

// ---------------------------------------------------------------------------
//	
//
bool OPNIF::Init(Bus* b, int intrport, int io, Scheduler* s, uint c, uint r)
{
	bus = b;
	scheduler = s;
	portio = io;
	opn.SetIntr(bus, intrport);
	
	if (!opn.Init(c, r, 0))
		return false;
	if (!SetRate(r, c))
		return false;
	prevtime = scheduler->GetTime();
	TimeEvent();
	return true;
}

// ---------------------------------------------------------------------------
//	EĐ[gݒ
//
bool OPNIF::SetRate(uint _clock, uint rate)
{
	clock = _clock;
	bool fm55k = false;
	if (rate == SOUND_55K)
	{
		fm55k = true;
		rate = 44100;
	}
	srate = rate;
	opn.SetReg(prescaler, 0);
	opn.SetRate(clock, srate, fm55k);
	return true;
}

// ---------------------------------------------------------------------------
//	
//
void OPNIF::Mix(FM::Sample* dest, int nsamples)
{
	if (enable)
		opn.Mix(dest, nsamples);
}

// ---------------------------------------------------------------------------
//	ʐݒ
//
static inline int ConvertVolume(int volume)
{
	return volume > -40 ? volume : -200;
}

void OPNIF::SetVolume(const Config* config)
{
	opn.SetVolumeFM(ConvertVolume(config->volfm));
	opn.SetVolumePSG(ConvertVolume(config->volssg));
#ifndef USE_OPN
	opn.SetVolumeADPCM(ConvertVolume(config->voladpcm));
	opn.SetVolumeRhythmTotal(ConvertVolume(config->volrhythm));
	opn.SetVolumeRhythm(0, ConvertVolume(config->volbd));
	opn.SetVolumeRhythm(1, ConvertVolume(config->volsd));
	opn.SetVolumeRhythm(2, ConvertVolume(config->voltop));
	opn.SetVolumeRhythm(3, ConvertVolume(config->volhh));
	opn.SetVolumeRhythm(4, ConvertVolume(config->voltom));
	opn.SetVolumeRhythm(5, ConvertVolume(config->volrim));
#endif
}

// ---------------------------------------------------------------------------
//	Reset
//
void OPNIF::Reset(uint, uint)
{
	memset(regs, 0, sizeof(regs));
	
	regs[0x29] = 0x1f;
	regs[0x110] = 0x1c;
	for (int i=0; i<3; i++)
		regs[0xb4+i] = regs[0x1b4+i] = 0xc0;

	
	opn.Reset();
	opn.SetIntrMask(true);
	prescaler = 0x2d;
}

// ---------------------------------------------------------------------------
//	荞
//
void OPNIF::OPNUnit::Intr(bool flag)
{
	bool prev = intrpending && intrenabled && bus;
	intrpending = flag;
	LOG3("OPN     :Interrupt %d %d %d\n", intrpending, intrenabled, !prev);
	if (intrpending && intrenabled && bus && !prev)
	{
		bus->Out(pintr, true);
	}
}

// ---------------------------------------------------------------------------
//	荞݋H
//
inline void OPNIF::OPNUnit::SetIntrMask(bool en)
{
	bool prev = intrpending && intrenabled && bus;
	intrenabled = en;
	if (intrpending && intrenabled && bus && !prev)
		bus->Out(pintr, true);
}

void OPNIF::SetIntrMask(uint a, uint intrmask)
{
//	LOG2("Intr enabled (%.2x)[%.2x]\n", a, intrmask);
	opn.SetIntrMask(!(imaskbit & intrmask));
}

// ---------------------------------------------------------------------------
//	SetRegisterIndex
//
void OPNIF::SetIndex0(uint a, uint data)
{
//	LOG2("Index0[%.2x] = %.2x\n", a, data);
	index0 = data;
	if (enable && (data & 0xfc) == 0x2c)
	{
		regs[0x2f] = 1;
		prescaler = data;
		opn.SetReg(data, 0);
	}
}

void OPNIF::SetIndex1(uint a, uint data)
{
//	LOG2("Index1[%.2x] = %.2x\n", a, data);
	index1 = data1 = data;
}

// ---------------------------------------------------------------------------
//	WriteRegister
//
void OPNIF::WriteData0(uint a, uint data)
{
//	LOG2("Write0[%.2x] = %.2x\n", a, data);
	if (enable)
	{
		LOG3("%.8x:OPN[0%.2x] = %.2x\n", scheduler->GetTime(), index0, data);
		TimeEvent(1);
		
		if (!opnamode)
		{
			if ((index0 & 0xf0) == 0x10) return;
			if (index0 == 0x22) data = 0;
			if (index0 == 0x28 && (data & 4)) return;
			if (index0 == 0x29) data = 3;
			if (index0 >= 0xb4) data = 0xc0;
		}
		regs[index0] = data;
		opn.SetReg(index0, data);
		if (index0 == 0x27)
		{
			UpdateTimer();
		}
	}
}

void OPNIF::WriteData1(uint a, uint data)
{
//	LOG2("Write1[%.2x] = %.2x\n", a, data);
#ifndef USE_OPN
	if (enable && opnamode)
	{
		LOG3("%.8x:OPN[1%.2x] = %.2x\n", scheduler->GetTime(), index1, data);
		if (index1 != 0x08 && index1 != 0x10)
			TimeEvent(1);
		data1 = data;
		regs[0x100 | index1] = data;
		opn.SetReg(0x100 | index1, data);
	}
#endif
}

// ---------------------------------------------------------------------------
//	ReadRegister
//
uint OPNIF::ReadData0(uint a)
{
	uint ret;
	if (!enable)
		ret = 0xff;
	else if ((index0 & 0xfe) == 0x0e)
		ret = bus->In(portio + (index0 & 1));
	else if (index0 == 0xff && !opnamode)
		ret = 0;
	else
		ret = opn.GetReg(index0);
//	LOG2("Read0 [%.2x] = %.2x\n", a, ret);
	return ret;
}

uint OPNIF::ReadData1(uint a)
{
	uint ret = 0xff;
#ifndef USE_OPN
	if (enable && opnamode)
	{
		if (index1 == 0x08)
			ret = opn.GetReg(0x100 | index1);
		else
			ret = data1;
	}
//	LOG3("Read1 [%.2x] = %.2x  (d1:%.2x)\n", a, ret, data1);
#endif
	return ret;
}

// ---------------------------------------------------------------------------
//	ReadStatus
//
uint OPNIF::ReadStatus(uint a)
{
	uint ret = enable ? opn.ReadStatus() : 0xff;
//	LOG2("status[%.2x] = %.2x\n", a, ret);
	return ret;
}

uint OPNIF::ReadStatusEx(uint a)
{
	uint ret = enable && opnamode ? opn.ReadStatusEx() : 0xff;
//	LOG2("statex[%.2x] = %.2x\n", a, ret);
	return ret;
}

// ---------------------------------------------------------------------------
//	^C}[XV
//
void OPNIF::UpdateTimer()
{
	static Scheduler::Event* ev;
	if (!ev || !ev->inst)
		LOG1("%.8x:UpdateTimer()", scheduler->GetTime());
	else
		LOG3("%.8x:UpdateTimer() [%.8x/%8d]", scheduler->GetTime(), ev->time, ev->count);

	scheduler->DelEvent(this);
	nextcount = opn.GetNextEvent();
	if (nextcount)
	{
		nextcount = (nextcount+9) / 10;
		ev = scheduler->AddEvent(nextcount, this, STATIC_CAST(TimeFunc, &OPNIF::TimeEvent), 0);
		LOG2(" [%.8x] - %8d\n", nextcount + scheduler->GetTime(), nextcount);
	}
	LOG0("\n");
}

// ---------------------------------------------------------------------------
//	^C}[
//
void OPNIF::TimeEvent(uint rtc)
{
	int currenttime = scheduler->GetTime();
	int diff = currenttime - prevtime;
	prevtime = currenttime;

	if (enable)
	{
		LOG3("%.8x:TimeEvent(%d) : diff:%d\n", currenttime, rtc, diff);
		
		if (sound)
			sound->Update();
		if (opn.Count(diff * 10) || !rtc)
			UpdateTimer();
	}
}

// ---------------------------------------------------------------------------
//	ԕۑ
//
uint OPNIF::GetStatusSize()
{
	if (enable)
		return sizeof(Status) + (opnamode ? 0x40000 : 0);
	else
		return 0;
}

bool OPNIF::SaveStatus(uint8* s)
{
	Status* st = (Status*) s;
	st->rev = ssrev;
	st->i0 = index0;
	st->i1 = index1;
	st->d0 = 0;
	st->d1 = data1;
	st->is = opn.IntrStat();
	memcpy(st->regs, regs, 0x200);
	if (opnamode)
		memcpy(s + sizeof(Status), opn.GetADPCMBuffer(), 0x40000);
	return true;
}

bool OPNIF::LoadStatus(const uint8* s)
{
	const Status* st = (const Status*) s;
	if (st->rev != ssrev)
		return false;

	prevtime = scheduler->GetTime();
	
	int i;
	for (i=8; i<=0x0a; i++)
		opn.SetReg(i, 0);
	for (i=0x40; i<0x4f; i++)
		opn.SetReg(i, 0x7f), opn.SetReg(i+0x100, 0x7f);

	for (i=0; i<0x10; i++)
		SetIndex0(0, i), WriteData0(0, st->regs[i]);

	opn.SetReg(0x10, 0xdf);
	
	for (i=11; i<0x28; i++)
		SetIndex0(0, i), WriteData0(0, st->regs[i]);

	SetIndex0(0, 0x29), WriteData0(0, st->regs[0x29]);
	
	for (i=0x30; i<0xa0; i++)
	{
		index0 = i, WriteData0(0, st->regs[i]);
		index1 = i, WriteData1(0, st->regs[i+0x100]);
	}
	for (i=0xb0; i<0xb7; i++)
	{
		index0 = i, WriteData0(0, st->regs[i]);
		index1 = i, WriteData1(0, st->regs[i+0x100]);
	}
	for (i=0; i<3; i++)
	{
		index0 = 0xa4 + i, WriteData0(0, st->regs[0xa4 + i]);
		index0 = 0xa0 + i, WriteData0(0, st->regs[0xa0 + i]);
		index0 = 0xac + i, WriteData0(0, st->regs[0xac + i]);
		index0 = 0xa8 + i, WriteData0(0, st->regs[0xa8 + i]);
		index1 = 0xa4 + i, WriteData1(0, st->regs[0x1a4 + i]);
		index1 = 0xa0 + i, WriteData1(0, st->regs[0x1a0 + i]);
		index1 = 0xac + i, WriteData1(0, st->regs[0x1ac + i]);
		index1 = 0xa8 + i, WriteData1(0, st->regs[0x1a8 + i]);
	}		
	for (index1=0x00; index1<0x08; index1++)
		WriteData1(0, st->regs[0x100 | index1]);
	for (index1=0x09; index1<0x0e; index1++)
		WriteData1(0, st->regs[0x100 | index1]);

	opn.SetIntrMask(!!(st->is & 1));
	opn.Intr(!!(st->is & 2));
	index0 = st->i0;
	index1 = st->i1;
	data1  = st->d1;

	if (opnamode)
		memcpy(opn.GetADPCMBuffer(), s + sizeof(Status), 0x40000);
	UpdateTimer();
	return true;
}

// ---------------------------------------------------------------------------
//	device description
//
const Device::Descriptor OPNIF::descriptor = { indef, outdef };

const Device::OutFuncPtr OPNIF::outdef[] = 
{
	STATIC_CAST(Device::OutFuncPtr, Reset),
	STATIC_CAST(Device::OutFuncPtr, SetIndex0),
	STATIC_CAST(Device::OutFuncPtr, SetIndex1),
	STATIC_CAST(Device::OutFuncPtr, WriteData0),
	STATIC_CAST(Device::OutFuncPtr, WriteData1),
	STATIC_CAST(Device::OutFuncPtr, SetIntrMask),
};

const Device::InFuncPtr OPNIF::indef[] = 
{
	STATIC_CAST(Device::InFuncPtr, ReadStatus),
	STATIC_CAST(Device::InFuncPtr, ReadStatusEx),
	STATIC_CAST(Device::InFuncPtr, ReadData0),
	STATIC_CAST(Device::InFuncPtr, ReadData1),
};
