// ---------------------------------------------------------------------------
//	M88 - PC-8801 Emulator
//	Copyright (C) cisc 1997, 1999.
// ---------------------------------------------------------------------------
//	$Id: opnif.h,v 1.9 1999/08/23 11:58:53 cisc Exp $

#ifndef PC88_OPNIF_H
#define PC88_OPNIF_H

#include "device.h"
#include "opna.h"

// ---------------------------------------------------------------------------

class PC88;
class Scheduler;

//#define USE_OPN

namespace PC8801
{
class Sound;
class Config;

// ---------------------------------------------------------------------------
//	88 p OPN Unit
//
class OPNIF : public Device
{
public:
	enum IDFunc
	{
		reset=0, setindex0, setindex1, writedata0, writedata1, setintrmask,
		readstatus=0, readstatusex, readdata0, readdata1,
	};

public:
	OPNIF(const ID& id, int imaskbit);
	~OPNIF();
	
	bool Init(Bus* bus, int intrport, int io, Scheduler* s, uint c, uint r);
	void SetSound(Sound* s) { sound = s; }
	
	uint GetStatusSize();
	bool SaveStatus(uint8* status);
	bool LoadStatus(const uint8* status);
	
	bool SetRate(uint c, uint r);
	void Mix(FM::Sample* buffer, int nsamples);
	void SetVolume(const Config* config);
	
	void Enable(bool en) { enable = en; }
	void SetOPNMode(bool _opna) { opnamode = _opna; }
	const uint8* GetRegs() { return regs; }
	void SetChannelMask(uint ch);
	
	void SetIntrMask(uint, uint intrmask);
	void Reset		(uint=0, uint=0);
	void SetIndex0	(uint, uint data);
	void SetIndex1	(uint, uint data);
	void WriteData0	(uint, uint data);
	void WriteData1	(uint, uint data);
	uint ReadData0	(uint);
	uint ReadData1	(uint);
	uint ReadStatus	(uint);
	uint ReadStatusEx(uint);
	
	void TimeEvent(uint=0);
	
	const Descriptor* GetDesc() const { return &descriptor; }

private:
	void UpdateTimer();

private:
	class OPNUnit : 
#ifndef USE_OPN
	public FM::OPNA
#else
	public FM::OPN
#endif
	{
	public:
		OPNUnit() : bus(0) {}
		~OPNUnit() {}
		void Intr(bool f);
		void SetIntr(Bus* b, int p) { bus = b, pintr = p; }
		void SetIntrMask(bool e);
		uint IntrStat() { return (intrenabled ? 1 : 0) | (intrpending ? 2 : 0); }

	private:
		Bus* bus;
		int pintr;
		bool intrenabled;
		bool intrpending;

		friend class OPNIF;
	};

	enum
	{
		ssrev = 1,
	};
	struct Status
	{
		uint8 rev;
		uint8 i0, i1, d0, d1;
		uint8 is;
		uint8 regs[0x200];
	};

	OPNUnit opn;
	Bus* bus;
	Scheduler* scheduler;
	Sound* sound;
	int32 clock;
	int32 nextcount;
	int imaskbit;
	int srate;
	int prevtime;
	int portio;

	bool opnamode;
	bool enable;
	
	uint index0;
	uint index1;
	uint data1;
	
	uint8 regs[0x200];

	static int prescaler;

	static const Descriptor descriptor;
	static const InFuncPtr  indef[];
	static const OutFuncPtr outdef[];
};


inline void OPNIF::SetChannelMask(uint ch)
{
	opn.SetChannelMask(ch);
}


}

#endif // PC88_OPNIF_H
