// ---------------------------------------------------------------------------
//	M88 - PC-8801 Emulator.
//	Copyright (C) cisc 1998, 1999.
// ---------------------------------------------------------------------------
//	$Id: wincore.cpp,v 1.23 1999/08/26 08:07:04 cisc Exp $

#include "headers.h"
#include <process.h>
#include "wincore.h"
#include "WinKeyIF.h"
#include "misc.h"
#include "pc88/config.h"
#include "status.h"
#include "device.h"
#include "version.h"
#include "file.h"
#include "pc88/diskmgr.h"
#include "extdev.h"

//#define LOGNAME "wincore"
#include "diag.h"

using namespace PC8801;

// ---------------------------------------------------------------------------
//	\z/
//
WinCore::WinCore()
: sound(DEV_ID('S', 'N', 'D', ' ')), mouse(DEV_ID('M', 'O', 'U', 'S'))
{
	hthread = 0;
	execcount = 0;
	extdev = 0;
}

WinCore::~WinCore()
{
	Cleanup();
}

// ---------------------------------------------------------------------------
//	
//
bool WinCore::Init
(WinUI* _ui, HWND hwnd, Draw* draw, DiskManager* disk, WinKeyIF* keyb)
{
	ui = _ui;

	if (!extdev)
	{
		extdev = new ExternalDevice[num_externals];
		if (!extdev)
			return false;
	}
	
	if (!PC88::Init(draw, disk))
		return false;

	if (!sound.Init(this, hwnd, 0, 3993600, 0))
		return false;

	if (!ConnectDevices(keyb))
		return false;

	if (!ConnectExternalDevices())
		return false;

	execute = false;
	shouldterminate = false;
	padenable = false;
	mouseenable = false;
	execcount = 0;
	guicount = 0;
	clock = 40;
	if (!hthread)
	{
		hthread = HANDLE(_beginthreadex(NULL, 0, ThreadEntry,
							reinterpret_cast<void*>(this), 0, &idthread));
		if (!hthread)
			return false;
	}
	return true;
}

// ---------------------------------------------------------------------------
//	n
//
bool WinCore::Cleanup()
{
	if (hthread)
	{
		shouldterminate = true;
		if (WAIT_TIMEOUT == WaitForSingleObject(hthread, 3000))
			TerminateThread(hthread, 0);
		CloseHandle(hthread);
		hthread = 0;
	}
	delete[] extdev; extdev = 0;
	return true;
}

// ---------------------------------------------------------------------------
//	Zbg
//
void WinCore::Reset()
{
	CriticalSection::Lock lock(cs_cpu);
	PC88::Reset();
}

// ---------------------------------------------------------------------------
//	Core Thread
//
uint WinCore::ThreadMain()
{
	DWORD time = GetTime();

	int effclock = 100;

	while (!shouldterminate)
	{
		if (execute)
		{
			if (!clock)
			{
				time = GetTime();
				DWORD ms;
				int eclk = 0;
				do
				{
					Execute(effclock, 500);
					eclk+=5;
					ms = GetTime() - time;
				} while (ms < 10);

				// ̎NbNvZ  A 10GHz? (^^;
				effclock = Min((Min(1000, eclk) * effclock / ms) + 1, 10000);
			}
			else
			{
				if (clock > 0)
				{
					int dt = GetTime() - time;
					int ms = Limit(dt, 50, 4);
					time += dt;								// time  I\莞
					Execute(clock, ms * 100);

					int idletime = time - GetTime() + 10;	// idletime  ]莞
//					LOG3("Time: %d  Ex: %d ms  It: %d ms\n", GetTime(), ms, Max(0, idletime));
					if (idletime > 0)
						Sleep(idletime);
				}
				else
					Execute(-clock, 500);
			}
		}
		else
		{
			Sleep(5);
			time = GetTime();
		}
	}
	return 0;
}

// ---------------------------------------------------------------------------
//	TuXbhJn_
//
uint __stdcall WinCore::ThreadEntry(void* arg)
{
	return reinterpret_cast<WinCore*>(arg)->ThreadMain();
}


// ---------------------------------------------------------------------------
//	botC[v
//	clock	bot̃NbN(0.1MHz)
//	time	s鎞 (0.01ms)
//
void WinCore::Execute(long clk, long time)
{
	CriticalSection::Lock lock(cs_cpu);
	execcount += clk * Proceed(time, clk);
}

// ---------------------------------------------------------------------------
//	sNbNJEg̒lԂAJE^Zbg
//
long WinCore::GetExecCount()
{
	long i = execcount;
	execcount = 0;
	return i;
}

// ---------------------------------------------------------------------------
//	s~邩H
//
void WinCore::Wait(bool dowait)
{
	CriticalSection::Lock lock(cs_cpu);
	execute = !dowait;
}

// ---------------------------------------------------------------------------
//	Ԃ擾
//
inline uint32 WinCore::GetTime()
{
	return timeGetTime();
}

// ---------------------------------------------------------------------------
//	ݒ𔽉f
//
void WinCore::ApplyConfig(PC8801::Config* cfg)
{
	config = *cfg;
	clock = cfg->clock;
	if (cfg->flags & PC8801::Config::fullspeed)
		clock = 0;
	if (cfg->flags & PC8801::Config::cpuburst)
		clock = -clock;
	PC88::ApplyConfig(cfg);
	sound.ApplyConfig(cfg);
	pad.ApplyConfig(cfg);
	mouse.ApplyConfig(cfg);

	EnablePad((cfg->flags & PC8801::Config::enablepad) != 0);
	if (padenable)
		cfg->flags &= ~PC8801::Config::enablemouse;
	EnableMouse((cfg->flags & PC8801::Config::enablemouse) != 0);
}

// ---------------------------------------------------------------------------
//	GUI gptOݒ
//
void WinCore::SetGUIFlag(bool usegui)
{
	if (usegui)
	{
		if (!guicount++)
			mouse.Enable(false);
	}
	else
	{
		if (!--guicount)
			mouse.Enable(mouseenable);
	}
}

// ---------------------------------------------------------------------------
//	Windows p̃foCXڑ
//
bool WinCore::ConnectDevices(WinKeyIF* keyb)
{
	const static Bus::Connector c_keyb[] = 
	{
		{ PC88::pres, Bus::portout, WinKeyIF::reset },
		{ PC88::vrtc, Bus::portout, WinKeyIF::vsync },
		{ 0x00, Bus::portin, WinKeyIF::in },
		{ 0x01, Bus::portin, WinKeyIF::in },
		{ 0x02, Bus::portin, WinKeyIF::in },
		{ 0x03, Bus::portin, WinKeyIF::in },
		{ 0x04, Bus::portin, WinKeyIF::in },
		{ 0x05, Bus::portin, WinKeyIF::in },
		{ 0x06, Bus::portin, WinKeyIF::in },
		{ 0x07, Bus::portin, WinKeyIF::in },
		{ 0x08, Bus::portin, WinKeyIF::in },
		{ 0x09, Bus::portin, WinKeyIF::in },
		{ 0x0a, Bus::portin, WinKeyIF::in },
		{ 0x0b, Bus::portin, WinKeyIF::in },
		{ 0x0c, Bus::portin, WinKeyIF::in },
		{ 0x0d, Bus::portin, WinKeyIF::in },
		{ 0x0e, Bus::portin, WinKeyIF::in },
		{ 0x0f, Bus::portin, WinKeyIF::in },
		{ 0, 0, 0 }
	};
	if (!bus1.Connect(keyb, c_keyb)) return false;

	const static Bus::Connector c_snd[] = 
	{
		{ 0x40,        Bus::portout, WinSound::out40 },
		{ PC88::pint1, Bus::portout, WinSound::updatecounter },
		{ 0, 0, 0 }
	};
	if (!bus1.Connect(&sound, c_snd)) return false;

	if (!mouse.Init(ui, this)) return false;
	return true;
}

// ---------------------------------------------------------------------------
//	PAD ڑ
//
bool WinCore::EnablePad(bool enable)
{
	if (padenable == enable)
		return true;
	padenable = enable;

	CriticalSection::Lock lock(cs_cpu);
	if (enable)
	{
		const static Bus::Connector c_pad[] =
		{
			{ PC88::popnio  , Bus::portin, WinJoyPad::getdir },
			{ PC88::popnio+1, Bus::portin, WinJoyPad::getbutton },
			{ PC88::vrtc,     Bus::portout, WinJoyPad::vsync },
			{ 0, 0, 0 }
		};
		if (!bus1.Connect(&pad, c_pad)) return false;
		pad.Init();
	}
	else
	{
		bus1.Disconnect(&pad);
	}
	return true;
}

// ---------------------------------------------------------------------------
//	}EXڑ
//
bool WinCore::EnableMouse(bool enable)
{
	if (mouseenable == enable)
		return true;
	mouseenable = enable;

	CriticalSection::Lock lock(cs_cpu);
	if (enable)
	{
		const static Bus::Connector c_mouse[] =
		{
			{ PC88::popnio  , Bus::portin, WinMouse::getmove },
			{ PC88::popnio+1, Bus::portin, WinMouse::getbutton },
			{ 0x40,           Bus::portout, WinMouse::strobe },
			{ PC88::vrtc,     Bus::portout, WinMouse::vsync },
			{ 0, 0, 0 }
		};
		if (!bus1.Connect(&mouse, c_mouse)) return false;
		ActivateMouse(true);
	}
	else
	{
		bus1.Disconnect(&mouse);
		ActivateMouse(false);
	}
	return true;
}

// ---------------------------------------------------------------------------
//	CPU1 ̎sO_v
//
bool WinCore::EnableCPU1Dump(bool f)
{
	CriticalSection::Lock lock(cs_cpu);
	return GetCPU1()->EnableDump(f);
}

// ---------------------------------------------------------------------------
//	Snapshot wb_[
//
enum
{
	ssmajor = 1, ssminor = 0,
};

struct SnapshotHeader
{
	char id[16];
	uint8 major, minor;

	int8 disk[2];
	uint datasize;
	Config::BASICMode basicmode;
	int16 clock;
	uint16 erambanks;
	uint16 cpumode;
	uint16 mainsubratio;
	uint flags;
	uint flag2;
};

//					 0123456789abcdef
#define SNAPSHOT_ID	"M88 SnapshotData"

// ---------------------------------------------------------------------------
//	XibvVbgۑ
//
bool WinCore::SaveShapshot(const char* filename)
{
	CriticalSection::Lock lock(cs_cpu);

	uint size = devlist.GetStatusSize();
	uint8* buf = new uint8[size];
	if (!buf)
		return false;
	memset(buf, 0, size);

	if (devlist.SaveStatus(buf))
	{
		SnapshotHeader ssh;
		memcpy(ssh.id, SNAPSHOT_ID, 16);

		ssh.major = ssmajor;
		ssh.minor = ssminor;
		ssh.datasize = size;
		ssh.basicmode = config.basicmode;
		ssh.clock = int16(config.clock);
		ssh.erambanks = uint16(config.erambanks);
		ssh.cpumode = int16(config.cpumode);
		ssh.mainsubratio = int16(config.mainsubratio);
		ssh.flags = config.flags;
		ssh.flag2 = config.flag2;
		for (uint i=0; i<2; i++)
			ssh.disk[i] = (int8) diskmgr->GetCurrentDisk(i);

		FileIO file;
		if (file.Open(filename, FileIO::create))
		{
			file.Write(&ssh, sizeof(ssh));
			file.Write(buf, size);
		}
	}
	delete[] buf;
	return true;
}

// ---------------------------------------------------------------------------
//	XibvVbg
//
bool WinCore::LoadShapshot(const char* filename, const char* diskname)
{
	CriticalSection::Lock lock(cs_cpu);

	FileIO file;
	if (!file.Open(filename, FileIO::readonly))
		return false;

	SnapshotHeader ssh;
	if (file.Read(&ssh, sizeof(ssh)) != sizeof(ssh))
		return false;
	if (memcmp(ssh.id, SNAPSHOT_ID, 16))
		return false;
	if (ssh.major != ssmajor || ssh.minor > ssminor)
		return false;

	// applyconfig
	const uint fl1a = Config::subcpucontrol | Config::fullspeed 
					| Config::enableopna	| Config::enablepcg
					| Config::fv15k 		| Config::cpuburst 
					| Config::cpuclockmode	| Config::digitalpalette
					| Config::opnona8		| Config::opnaona8
					| Config::enablewait;
	const uint fl2a = Config::disableopn44;

	config.flags = (config.flags & ~fl1a) | (ssh.flags & fl1a);
	config.flag2 = (config.flag2 & ~fl2a) | (ssh.flag2 & fl2a);
	config.basicmode = ssh.basicmode;
	config.clock = ssh.clock;
	config.erambanks = ssh.erambanks;
	config.cpumode = ssh.cpumode;
	config.mainsubratio = ssh.mainsubratio;
	ApplyConfig(&config);
	
	// Reset
	PC88::Reset();

	// ǂݍ

	uint8* buf = new uint8[ssh.datasize];
	bool r = false;

	if (buf)
	{
		uint rdsize = file.Read(buf, ssh.datasize);
		if (rdsize == ssh.datasize)
		{
			r = devlist.LoadStatus(buf);
			if (r && diskname)
			{
				for (uint i=0; i<2; i++)
				{
					diskmgr->Unmount(i);
					diskmgr->Mount(i, diskname, false, ssh.disk[i], false);
				}
			}
			if (!r)
			{
				statusdisplay.Show(70, 3000, "o[WقȂ܂");
				PC88::Reset();
			}
		}
		delete[] buf;
	}
	return r;
}

bool WinCore::ConnectExternalDevices()
{
	FileFinder ff;
	extern char m88dir[MAX_PATH];
	char buf[MAX_PATH];
	strncpy(buf, m88dir, MAX_PATH);
	strncat(buf, "*.m88", MAX_PATH);

	if (ff.FindFile(buf))
	{
		int i=0;
		while (ff.FindNext() && i < num_externals)
		{
			if (extdev[i].Init(ff.GetFileName(), this, &bus1, dmac, &sound))
			{
				devlist.Add(&extdev[i]);
				i++;
			}
		}
	}
	return true;
}

